﻿using Eingabeverwaltung.Metamodell;
using System.Collections.Generic;
using System;

namespace Ueberpruefung
{
    public static class CheckHelper
    {
        public static Tree substitute(Tree tree, Substitution sub) => substitute(tree, new List<Substitution>() { sub });
        public static Tree substitute(Tree tree, List<Substitution> subs)
        {
            foreach(Substitution sub in subs)
            {
                if (tree as TreeVariable != null)
                {
                    TreeVariable workingTree = tree as TreeVariable;
                    if (workingTree.variable == sub.variable) tree = sub.tree.Copy();
                }
                else
                {
                    if (tree as ComposedTree != null)
                    {
                        ComposedTree workingTree = tree as ComposedTree;
                        List<Tree> substitutedChildren = new List<Tree>();
                        for (int i = 0; i < workingTree.children.Count; i++)
                            substitutedChildren.Add(substitute(workingTree.children[i], sub));
                        tree = new ComposedTree(workingTree.operation, substitutedChildren);
                    }
                    else throw new Exception("Found a tree that is wether TreeVariable nor ComposedTree: " + tree.ToString());
                }
            }
            return tree;
        }

        /// <summary>
        /// finds substitution from fromEq to toEq. Splits the found substitutions in substitutions with composed trees (subs) and
        /// substitutions with tree variable (renamings). Subs is null, if result is false.
        /// </summary>
        /// <param name="fromEq"></param>
        /// <param name="toEq"></param>
        /// <param name="subs"></param>
        /// <param name="renamings"></param>
        public static bool getSubsAndRenamings(Equation fromEq, Equation toEq, out List<Substitution> subs, out List<Substitution> renamings)
        {
            subs = findSubstitution(fromEq, toEq, new List<Variable>(), new List<Variable>());
            renamings = new List<Substitution>();
            // split in renamings and real substitutions
            if (subs == null)
                return false;
            else
                for (int i = 0; i < subs.Count; i++)
                    if (subs[i].tree is TreeVariable)
                    {
                        renamings.Add(subs[i]);
                        subs.RemoveAt(i); i--;
                    }
            return true;
        }

        /// <summary>
        /// Same as findSubstitution for Trees, but this is for Equation.
        /// It tries substitution for equal direction first, if fails tries substitution for crossed direction.
        /// </summary>
        /// <param name="forEq"></param>
        /// <param name="toEq"></param>
        /// <param name="forFixed">Fixed variables in forEq. Input empty List here if you want to ignore that variables are fixed.</param>
        /// <param name="toFixed">Fixed variables in toEq. Input empty List here if you don't want to substitute fixed variables at all.</param>
        /// <returns></returns>
        public static List<Substitution> findSubstitution(Equation forEq, Equation toEq, List<Variable> forFixed, List<Variable> toFixed)
        {
            List<Substitution> findSubsOneDirection(Equation forEq, Equation toEq, List<Variable> forFixed, List<Variable> toFixed, bool swapped)
            {
                List<Substitution> result1 = null, result2 = null;
                if (!swapped)
                {
                    result1 = findSubstitution(forEq.leftTree , toEq.leftTree , forFixed, toFixed);
                    result2 = findSubstitution(forEq.rightTree, toEq.rightTree, forFixed, toFixed);
                }
                else
                {
                    result1 = findSubstitution(forEq.leftTree, toEq.rightTree, forFixed, toFixed);
                    result2 = findSubstitution(forEq.rightTree, toEq.leftTree, forFixed, toFixed);
                }
                return unionSubstitutions(result1, result2);
            }
            List<Substitution> result = findSubsOneDirection(forEq, toEq, forFixed, toFixed, false);
            if (result == null) result = findSubsOneDirection(forEq, toEq, forFixed, toFixed, true);
            return result;
        }

        /// <summary>
        /// Same as findSubstitution(forTree,toTree), but checks to not substitute fixed variables,
        /// except it is a renaming to another fixed variable with same name.
        /// </summary>
        /// <param name="forTree"></param>
        /// <param name="toTree"></param>
        /// <param name="forFixed">Fixed variables in forTree. Input empty List here if you want to ignore that variables are fixed.</param>
        /// <param name="toFixed">Fixed variables in toTree. Input empty List here if you don't want to substitute fixed variables at all.</param>
        /// <param name="allowFixedRenaming"> Determines if renaming of fixed variables to a fixed Variable with another name is allowed.</param>
        /// <returns></returns>
        public static List<Substitution> findSubstitution(Tree forTree, Tree toTree, List<Variable> forFixed, List<Variable> toFixed, bool allowFixedRenaming = false)
        {
            List<Substitution> result = findSubstitution(forTree, toTree);
            if (result != null)
                foreach(Substitution sub in result)
                    if (forFixed.Contains(sub.variable))
                        if (sub.tree is TreeVariable tv && toFixed.Contains(tv.variable) && (allowFixedRenaming || sub.variable.name == tv.variable.name)) { }
                        else return null; // substitution not possible due to fixedVars
            return result;
        }

        /// <summary>
        /// Finds a List of Substitutions for the <paramref name="forTree"/>, so if you use this Substitutions 
        /// on the <paramref name="forTree"/> you will get the <paramref name="toTree"/>.
        /// 
        /// </summary>
        /// <remarks> 
        /// Assumes all variables are allquantified. 
        /// Use findSubstitution with Lists of fixed variables, if the substitutions should consider fixed variables
        /// </remarks>
        /// <param name="forTree"></param>
        /// <param name="toTree"></param>
        /// <returns>
        /// Returns an empty List if forTree.Equals(toTree).
        /// Returns null if such a Substitution not exists.
        /// </returns>
        public static List<Substitution> findSubstitution(Tree forTree, Tree toTree)
        {
            if (forTree as TreeVariable != null)
                return new List<Substitution>() { new Substitution(((TreeVariable)forTree).variable, toTree) };
            else if (forTree as ComposedTree != null && toTree as ComposedTree != null)
            {
                ComposedTree forCTree = forTree as ComposedTree;
                ComposedTree toCTree = toTree as ComposedTree;
                if (forCTree.operation == toCTree.operation)
                {
                    List<Substitution> result = new List<Substitution>();
                    for (int forChild = 0; forChild < forCTree.children.Count; forChild++)
                    {
                        List<Substitution> childResult = findSubstitution(forCTree.children[forChild], toCTree.children[forChild]);
                        if (unionSubstitutions(result, childResult) == null) return null;
                    }
                    return result;
                }
                return null;
            }
            else if (forTree as ComposedTree != null && toTree as TreeVariable != null) return null;
            else throw new Exception("Found a tree that is wether TreeVariable nor ComposedTree: forTree: " + forTree.ToString() + ", toTree: " + toTree.ToString());            
        }

        /// <summary>
        /// Given two substitutions, result is the union of them.
        /// Returns null if one argument is null or the substitutions are incompatible.
        /// @pre: subs1 and subs2 are correct substitutions or null
        /// @post: result is correct substitution and result = @post.subs1 and @post.subs2 = @pre.subs2-@pre.subs1 or result = null
        /// </summary>
        /// <param name="subs1"></param>
        /// <param name="subs2"></param>
        /// <returns></returns>
        public static List<Substitution> unionSubstitutions(List<Substitution> subs1, List<Substitution> subs2)
        {
            if (subs1 == null || subs2 == null) return null;
            foreach (Substitution sub in subs1)
                for (int sub2 = 0; sub2 < subs2.Count; sub2++)
                {
                    if (sub.variable == subs2[sub2].variable)
                    {
                        if (!sub.tree.Equals(subs2[sub2].tree)) return null;
                        subs2.RemoveAt(sub2);
                        sub2--;
                    }
                }
            subs1.AddRange(subs2);
            return subs1;
        }

        /// <summary>
        /// Searches the subTree in the oldTree and replace it with newSubTree.
        /// Returns a new Tree or null if not able to find.
        /// </summary>
        /// <param name="oldTree"></param>
        /// <param name="subTree"></param>
        /// <param name="newSubTree"></param>
        /// <returns></returns>
        public static Tree transform(Tree oldTree, Tree subTree, Tree newSubTree, int matchingID = 1)
         => transform(oldTree, subTree, newSubTree, matchingID, new int[] { 0 });
        private static Tree transform(Tree oldTree, Tree subTree, Tree newSubTree, int matchingID, int[] foundMatchings )
        {
            Tree tree = oldTree.Copy();
            if (tree.Equals(subTree)) 
                if (++foundMatchings[0] == matchingID) return newSubTree;
                else return null;
            if (tree as TreeVariable != null) return null;
            if (tree as ComposedTree != null)
            {
                ComposedTree cTree = tree as ComposedTree;
                bool foundSubTree = false;
                for (int childNr = 0; childNr < cTree.children.Count; childNr++)
                {
                    Tree transformedSChild = transform(cTree.children[childNr], subTree, newSubTree, matchingID, foundMatchings);                    
                    if (transformedSChild != null)
                    {
                        cTree.children[childNr] = transformedSChild;
                        foundSubTree = true;
                        break;
                    }
                }
                if (foundSubTree) return cTree;
                else return null;
            }
            else throw new Exception("Found a tree that is wether TreeVariable nor ComposedTree: " + tree.ToString());
        }
    }
}
